/*
 *  Audit helper functions used throughout shadow-utils
 *
 *  Copyright (C) 2005, Red Hat, Inc.
 *
 *  Redistributable under the same terms as the other elements of
 *  shadow-utils v4.0.3.
 */

#include <config.h>

#ifdef WITH_AUDIT
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <libaudit.h>
#include <errno.h>
#include <stdio.h>
#include <sys/utsname.h>
#include <string.h>
int audit_fd;
#endif


/*
 * This function checks to see if we are running on an old kernel. If so 
 * we return 1. If we are on a good kernel, 0 is returned.
 */
#ifdef WITH_AUDIT
static int check_kernel_version(void)
{
	struct utsname u;
	char *kern, *rel;
	const char *bad_ver = "2.6.9-11.EL";
	unsigned int release, len;

	if (uname(&u))
		return 1;

	// Do the quick decision for -11
	len = strlen(bad_ver);
	if (strncmp(u.release, bad_ver, len) == 0) {
		/* The CAPP cert kernel has a '.' after EL */
		if (u.release[len] != '.')
			return 1;
	}

	// if not the 2.6.9 kernel, should be ok
	kern = strtok((char *)u.release, "-");
	if (kern == NULL)
		return 1;
	if (strcmp(kern, "2.6.9"))
		return 0;

	// now get piece to left of '.'
	rel = strtok(NULL, ".");
	if (rel == NULL) 
		return 1;

	// if GA kernel, don't log
	errno = 0;
	release = strtoul(rel, NULL, 10);
	if (errno)
		return 1;
	// 11.EL was checked for at the beginning. We need an exception
	// here since the CAPP cert kernel is in the -11 series, but it 
	// has correct and working audit system.
	if (release < 11)
		return 1;

	// U2 or CAPP cert kernel, we are safe.
	return 0;
}
#endif

void audit_help_open(void)
{
#ifdef WITH_AUDIT
	if (check_kernel_version()) {
		audit_fd = -1;
		return;
	}

	audit_fd = audit_open();
	if (audit_fd < 0) {
		/* You get ECONNREFUSED only when the kernel doesn't have
		 * audit compiled in. */
		if (errno == ECONNREFUSED)
			return;
		fprintf(stderr, "Cannot open audit interface - aborting.\n");
		exit(1);
	}
#endif
}

/*
 * This function will log a message to the audit system using a predefined
 * message format. Parameter usage is as follows:
 *
 * type - type of message: AUDIT_USER_CHAUTHTOK for changing any account 
 *	  attributes.
 * pgname - program's name
 * op  -  operation. "adding user", "changing finger info", "deleting group"
 * name - user's account or group name. If not available use NULL.
 * id  -  uid or gid that the operation is being performed on. This is used
 *	  only when user is NULL.
 * result - 1 is "success" and 0 is "failed"
 */
void audit_logger(int type, const char *pgname, const char *op, 
	const char *name, unsigned int id, int result)
{
#ifdef WITH_AUDIT
        if (audit_fd < 0)
                return;
        else {
		char buf[PATH_MAX];
		const char *success;

		if (result)
			success = "success";
		else
			success = "failed";

		if (name)
			snprintf(buf, sizeof(buf), "%s: op=%s acct=%s res=%s",
				pgname, op, name, success);
		else
			snprintf(buf, sizeof(buf), "%s: op=%s id=%u res=%s",
				pgname, op, id, success);

		audit_send_user_message(audit_fd, type, buf);
        }
#endif
}

